#pragma once

#include <ReferenceCounter.h>
#include <Vec3.h>
#include <Vec2.h>
#include <Writer.h>

#include "Effect.h"
#include "Inventory.h"
#include "NetworkResponse.h"

class EntityType;
class Dimension;
class Game;
class ItemSkill;

class ActionTarget
{
private:
    Framework::Vec3<int> blockPos;
    Direction targetBlockSide;
    int entityId;

public:
    ActionTarget( Framework::Vec3<int> blockPos, Direction blockSide );
    ActionTarget( int entityId );

    void applyItemSkillOnTarget( Entity* zActor, ItemSkill* zItemSkill, Item* zUsedItem, Game* zGame );
    void placeBlock( Entity* zActor, Item* zItem, Game* zGame );
    void save( Framework::StreamWriter* zWriter ) const;
};

class Entity : public Inventory
{
protected:
    float maxHP;
    float currentHP;
    float stamina;
    float maxStamina;
    float hunger;
    float maxHunger;
    float thirst;
    float maxThirst;
    float targetDistanceLimit;
    Framework::Vec3<float> speed;
    Framework::Vec3<float> faceDir;
    Framework::Vec3<float> faceOffset;
    Framework::RCArray<ItemSkill> skills;
    ActionTarget* target;
    const EntityType* zEntityType;
    int currentDimensionId;
    bool removed;
    float gravityMultiplier;
    int id;

    virtual void onDeath();
    virtual void useItem( const ItemType* zType, Item* zItem, Game* zGame );
    Entity( const EntityType* zType, Framework::Vec3<float> location, int dimensionId, int entityId );

public:
    void prepareTick( const Dimension* zDimension, Game* zGame );
    virtual void tick( const Dimension* zDimension, Game* zGame );

    virtual void api( Framework::StreamReader* zRequest, NetworkResponse* zResponse );

    virtual void onFall( float collisionSpeed );
    void setPosition( Framework::Vec3<float> pos );
    float getMaxHP() const;
    float getCurrentHP() const;
    float getStamina() const;
    float getMaxStamina() const;
    float getHunger() const;
    float getMaxHunger() const;
    float getThirst() const;
    float getMaxThirst() const;
    Framework::Vec3<float> getSpeed() const;
    Framework::Vec3<float> getFaceDir() const;
    Framework::Vec3<float> getPosition() const;
    float getGravityMultiplier() const;
    int getCurrentDimensionId() const;
    bool isRemoved() const;

    const EntityType* zType() const;
    const ActionTarget* zTarget() const;
    int getId() const;

    friend Effect;
    friend EntityType;
};